@@ -1,9 +1,8 @@
-use anyhow::{bail, Result};
+use anyhow::{Result, bail};
 use std::io::{BufRead, Write};
-use std::ops::Range;
 
-use crate::bounds::{BoundOrFiller, Side};
-use crate::cut_str::cut_str;
+use crate::bounds::{BoundOrFiller, Side, UserBoundsTrait};
+use crate::cut_str::read_and_cut_str;
 use crate::options::Opt;
 use crate::read_utils::read_line_with_eol;
 
@@ -13,11 +12,11 @@ fn cut_lines_forward_only<A: BufRead, B: Write>(
     opt: &Opt,
 ) -> Result<()> {
     let mut line_buf = String::with_capacity(1024);
-    let mut line_idx = 0;
+    let mut line_idx = usize::MAX;
     let mut bounds_idx = 0; // keep track of which bounds have been used
     let mut add_newline_next = false;
     while let Some(line) = read_line_with_eol(stdin, &mut line_buf, opt.eol) {
-        line_idx += 1;
+        line_idx = line_idx.wrapping_add(1); // use wrapping add so we can start at 0
 
         let line = line?;
         let line: &str = line.as_ref();
@@ -26,15 +25,15 @@ fn cut_lines_forward_only<A: BufRead, B: Write>(
         // Print the matching fields. Fields are ordered but can still be
         // duplicated, e.g. 1-2,2,3 , so we may have to print the same
         // line multiple times
-        while bounds_idx < opt.bounds.0.len() {
-            let bof = opt.bounds.0.get(bounds_idx).unwrap();
+        while bounds_idx < opt.bounds.len() {
+            let bof = opt.bounds.get(bounds_idx).unwrap();
 
             let b = match bof {
                 BoundOrFiller::Filler(f) => {
-                    stdout.write_all(f.as_bytes())?;
+                    stdout.write_all(f)?;
                     bounds_idx += 1;
 
-                    if opt.join && bounds_idx != opt.bounds.0.len() {
+                    if opt.join && bounds_idx != opt.bounds.len() {
                         stdout.write_all(&[opt.eol as u8])?;
                     }
 
@@ -51,13 +50,13 @@ fn cut_lines_forward_only<A: BufRead, B: Write>(
                 stdout.write_all(line.as_bytes())?;
                 add_newline_next = true;
 
-                if b.r == Side::Some(line_idx) {
+                if b.r().value_unchecked() == line_idx {
                     // we exhausted the use of that bound, move on
                     bounds_idx += 1;
                     add_newline_next = false;
 
                     // if opt.join and it was not the last matching bound
-                    if opt.join && bounds_idx != opt.bounds.0.len() {
+                    if opt.join && bounds_idx != opt.bounds.len() {
                         stdout.write_all(&[opt.eol as u8])?;
                     }
 
@@ -68,18 +67,19 @@ fn cut_lines_forward_only<A: BufRead, B: Write>(
             break; // nothing matched, let's go to the next line
         }
 
-        if bounds_idx == opt.bounds.0.len() {
+        if bounds_idx == opt.bounds.len() {
             // no need to read the rest, we don't have other bounds to test
             break;
         }
     }
 
     // Output is finished. Did we output every bound?
-    if let Some(BoundOrFiller::Bound(b)) = opt.bounds.0.get(bounds_idx) {
-        if b.r != Side::Continue {
-            // not good, we still have bounds to print but the input is exhausted
-            bail!("Out of bounds: {}", b);
-        }
+    if let Some(BoundOrFiller::Bound(b)) = opt.bounds.get(bounds_idx)
+    // we can ignore the sign because this is forward only
+        && b.r().value_unchecked() != Side::max_right()
+    {
+        // not good, we still have bounds to print but the input is exhausted
+        bail!("Out of bounds: {}", b);
     }
 
     stdout.write_all(&[opt.eol as u8])?;
@@ -88,25 +88,15 @@ fn cut_lines_forward_only<A: BufRead, B: Write>(
 }
 
 fn cut_lines<A: BufRead, B: Write>(stdin: &mut A, stdout: &mut B, opt: &Opt) -> Result<()> {
-    let mut buffer: Vec<u8> = Vec::with_capacity(32 * 1024);
-    stdin.read_to_end(&mut buffer)?;
-    let buffer_as_str = std::str::from_utf8(&buffer)?;
-    let mut bounds_as_ranges: Vec<Range<usize>> = Vec::with_capacity(100);
-    let mut compressed_line_buf = String::new();
+    unsafe {
+        let mutable_ptr = opt as *const Opt as *mut Opt;
+        (*mutable_ptr).read_to_end = true;
+    }
 
-    let buffer_as_str = buffer_as_str
-        .strip_suffix(opt.eol as u8 as char)
-        .unwrap_or(buffer_as_str);
+    // Just use read_and_cut_str, we're cutting a (big) string whose delimiter is newline
+    read_and_cut_str(stdin, stdout, opt)?;
 
-    // Just use cut_str, we're cutting a (big) string whose delimiter is newline
-    cut_str(
-        buffer_as_str,
-        opt,
-        stdout,
-        &mut bounds_as_ranges,
-        &mut compressed_line_buf,
-        &[opt.eol as u8],
-    )
+    Ok(())
 }
 
 pub fn read_and_cut_lines<A: BufRead, B: Write>(
@@ -131,8 +121,10 @@ pub fn read_and_cut_lines<A: BufRead, B: Write>(
 
 #[cfg(test)]
 mod tests {
+    use std::str::FromStr;
+
     use crate::{
-        bounds::{BoundsType, UserBounds, UserBoundsList},
+        bounds::{BoundsType, UserBoundsList},
         options::EOL,
     };
 
@@ -141,46 +133,16 @@ mod tests {
     fn make_lines_opt() -> Opt {
         Opt {
             bounds_type: BoundsType::Lines,
-            delimiter: String::from("\n"),
+            delimiter: "\n".into(),
             join: true,
             ..Opt::default()
         }
     }
 
-    const BOF_F1: BoundOrFiller = BoundOrFiller::Bound(UserBounds {
-        l: Side::Some(1),
-        r: Side::Some(1),
-    });
-
-    const BOF_F2: BoundOrFiller = BoundOrFiller::Bound(UserBounds {
-        l: Side::Some(2),
-        r: Side::Some(2),
-    });
-
-    const BOF_F3: BoundOrFiller = BoundOrFiller::Bound(UserBounds {
-        l: Side::Some(3),
-        r: Side::Some(3),
-    });
-
-    const BOF_R2_3: BoundOrFiller = BoundOrFiller::Bound(UserBounds {
-        l: Side::Some(2),
-        r: Side::Some(3),
-    });
-
-    const BOF_NEG1: BoundOrFiller = BoundOrFiller::Bound(UserBounds {
-        l: Side::Some(-1),
-        r: Side::Some(-1),
-    });
-
-    const BOF_F1_TO_END: BoundOrFiller = BoundOrFiller::Bound(UserBounds {
-        l: Side::Some(1),
-        r: Side::Continue,
-    });
-
     #[test]
     fn fwd_cut_one_field() {
         let mut opt = make_lines_opt();
-        opt.bounds = UserBoundsList(vec![BOF_F1]);
+        opt.bounds = UserBoundsList::from_str("1").unwrap();
 
         let mut input = b"a\nb".as_slice();
         let mut output = Vec::with_capacity(100);
@@ -191,7 +153,7 @@ mod tests {
     #[test]
     fn fwd_cut_multiple_fields() {
         let mut opt = make_lines_opt();
-        opt.bounds = UserBoundsList(vec![BOF_F1, BOF_F2]);
+        opt.bounds = UserBoundsList::from_str("1:2").unwrap();
 
         let mut input = b"a\nb".as_slice();
         let mut output = Vec::with_capacity(100);
@@ -202,7 +164,7 @@ mod tests {
     #[test]
     fn fwd_support_ranges() {
         let mut opt = make_lines_opt();
-        opt.bounds = UserBoundsList(vec![BOF_F1, BOF_R2_3]);
+        opt.bounds = UserBoundsList::from_str("1,2:3").unwrap();
 
         let mut input = b"a\nb\nc".as_slice();
         let mut output = Vec::with_capacity(100);
@@ -213,7 +175,7 @@ mod tests {
     #[test]
     fn fwd_supports_no_join() {
         let mut opt = make_lines_opt();
-        opt.bounds = UserBoundsList(vec![BOF_F1, BOF_F3]);
+        opt.bounds = UserBoundsList::from_str("1,3").unwrap();
         opt.join = false;
 
         let mut input = b"a\nb\nc".as_slice();
@@ -225,7 +187,7 @@ mod tests {
     #[test]
     fn fwd_supports_no_right_bound() {
         let mut opt = make_lines_opt();
-        opt.bounds = UserBoundsList(vec![BOF_F1_TO_END]);
+        opt.bounds = UserBoundsList::from_str("1:").unwrap();
 
         let mut input = b"a\nb".as_slice();
         let mut output = Vec::with_capacity(100);
@@ -236,7 +198,7 @@ mod tests {
     #[test]
     fn fwd_handle_out_of_bounds() {
         let mut opt = make_lines_opt();
-        opt.bounds = UserBoundsList(vec![BOF_F3]);
+        opt.bounds = UserBoundsList::from_str("3").unwrap();
         opt.join = true;
 
         let mut input = b"a\nb".as_slice();
@@ -248,7 +210,7 @@ mod tests {
     #[test]
     fn fwd_ignore_last_empty() {
         let mut opt = make_lines_opt();
-        opt.bounds = UserBoundsList(vec![BOF_F3]);
+        opt.bounds = UserBoundsList::from_str("3").unwrap();
 
         let mut input1 = b"a\nb".as_slice();
         let mut input2 = b"a\nb\n".as_slice();
@@ -266,7 +228,7 @@ mod tests {
     #[test]
     fn cut_lines_handle_negative_idx() {
         let mut opt = make_lines_opt();
-        opt.bounds = UserBoundsList(vec![BOF_NEG1]);
+        opt.bounds = UserBoundsList::from_str("-1").unwrap();
 
         let mut input = b"a\nb".as_slice();
         let mut output = Vec::with_capacity(100);
@@ -277,7 +239,7 @@ mod tests {
     #[test]
     fn cut_lines_ignore_last_empty_when_using_positive_idx() {
         let mut opt = make_lines_opt();
-        opt.bounds = UserBoundsList(vec![BOF_F3]);
+        opt.bounds = UserBoundsList::from_str("3").unwrap();
 
         let mut input1 = b"a\nb".as_slice();
         let mut input2 = b"a\nb\n".as_slice();
@@ -295,7 +257,7 @@ mod tests {
     #[test]
     fn cut_lines_ignore_last_empty_when_using_negative_idx() {
         let mut opt = make_lines_opt();
-        opt.bounds = UserBoundsList(vec![BOF_NEG1]);
+        opt.bounds = UserBoundsList::from_str("-1").unwrap();
 
         let mut input1 = b"a\nb".as_slice();
         let mut input2 = b"a\nb\n".as_slice();
@@ -313,9 +275,9 @@ mod tests {
     #[test]
     fn fwd_cut_zero_delimited() {
         let mut opt = make_lines_opt();
-        opt.bounds = UserBoundsList(vec![BOF_F1]);
+        opt.bounds = UserBoundsList::from_str("1").unwrap();
         opt.eol = EOL::Zero;
-        opt.delimiter = String::from("\0");
+        opt.delimiter = "\0".into();
 
         let mut input = b"a\0b".as_slice();
         let mut output = Vec::new();
@@ -326,13 +288,27 @@ mod tests {
     #[test]
     fn cut_lines_zero_delimited() {
         let mut opt = make_lines_opt();
-        opt.bounds = UserBoundsList(vec![BOF_F1]);
+        opt.bounds = UserBoundsList::from_str("1").unwrap();
         opt.eol = EOL::Zero;
-        opt.delimiter = String::from("\0");
+        opt.delimiter = "\0".into();
 
         let mut input = b"a\0b".as_slice();
         let mut output = Vec::new();
         cut_lines(&mut input, &mut output, &opt).unwrap();
         assert_eq!(output, b"a\0");
     }
+
+    #[test]
+    fn cut_lines_and_compress_delimiter() {
+        // compress is similar to greedy, but while greedy
+        // preserve the original delimiters in its output
+        // (in particular when outputting a range), instead
+        // "compress" first merge the delimiters, then cut.
+        let opt: Opt = "-l 1:2 -p".parse().unwrap();
+
+        let mut input = b"a\n\nb".as_slice();
+        let mut output = Vec::with_capacity(100);
+        cut_lines(&mut input, &mut output, &opt).unwrap();
+        assert_eq!(output, b"a\nb\n");
+    }
 }
