@@ -1,14 +1,95 @@
 use assert_cmd::Command;
 
 #[test]
-fn it_echo_non_delimited_line() {
+fn it_display_short_help_when_run_without_arguments() {
     let mut cmd = Command::cargo_bin(env!("CARGO_PKG_NAME")).unwrap();
 
-    let assert = cmd.write_stdin("foobar").assert();
+    let assert = cmd.assert();
+
+    assert.success().stdout(predicates::str::starts_with("tuc"));
+}
+
+#[test]
+fn it_display_help_when_requested() {
+    let mut cmd = Command::cargo_bin(env!("CARGO_PKG_NAME")).unwrap();
+
+    let assert = cmd.args(["-h"]).assert();
+    assert.success().stdout(predicates::str::starts_with("tuc"));
+
+    let assert = cmd.args(["--help"]).assert();
+    assert.success().stdout(predicates::str::starts_with("tuc"));
+}
+
+#[test]
+fn it_display_the_version_when_requested() {
+    use bstr::ByteSlice;
+    use predicates::prelude::*;
+
+    let mut cmd = Command::cargo_bin(env!("CARGO_PKG_NAME")).unwrap();
+
+    let expected_version_line = format!("tuc {}\n", env!("CARGO_PKG_VERSION"));
+
+    let assert = cmd.args(["-V"]).assert();
+    assert.success().stdout(predicate::function(|x: &[u8]| {
+        x.to_str_lossy() == expected_version_line
+    }));
+
+    let mut cmd = Command::cargo_bin(env!("CARGO_PKG_NAME")).unwrap();
+
+    let assert = cmd.args(["--version"]).assert();
+    assert.success().stdout(predicate::function(|x: &[u8]| {
+        x.to_str_lossy() == expected_version_line
+    }));
+}
+
+#[test]
+fn it_echo_first_field_in_non_delimited_line() {
+    let mut cmd = Command::cargo_bin(env!("CARGO_PKG_NAME")).unwrap();
+
+    let assert = cmd
+        .args(["-d", "/", "-f", "1"])
+        .write_stdin("foobar")
+        .assert();
 
     assert.success().stdout("foobar\n");
 }
 
+#[test]
+fn it_fails_on_out_of_bound_fields_in_non_delimited_line() {
+    let mut cmd = Command::cargo_bin(env!("CARGO_PKG_NAME")).unwrap();
+
+    let assert = cmd
+        .args(["-d", "/", "-f", "2"])
+        .write_stdin("foobar")
+        .assert();
+
+    assert.failure().stderr("Error: Out of bounds: 2\n");
+}
+
+#[test]
+fn it_use_fallbacks_in_non_delimited_line() {
+    let mut cmd = Command::cargo_bin(env!("CARGO_PKG_NAME")).unwrap();
+
+    let assert = cmd
+        .args(["-d", "/", "-f", "2", "--fallback-oob", "ok"])
+        .write_stdin("foobar")
+        .assert();
+
+    assert.success().stdout("ok\n");
+}
+
+#[test]
+fn it_accept_empty_fallbacks() {
+    let mut cmd = Command::cargo_bin(env!("CARGO_PKG_NAME")).unwrap();
+
+    let assert = cmd
+        .args(["-d", "/", "-f", "1,5,3", "--fallback-oob="])
+        .write_stdin("foobar/baz/what")
+        .assert();
+
+    assert.success().stdout("foobarwhat\n");
+}
+
 #[test]
 fn it_skips_non_delimited_line_when_requested() {
     let mut cmd = Command::cargo_bin(env!("CARGO_PKG_NAME")).unwrap();
@@ -42,16 +123,28 @@ fn it_cut_consecutive_delimiters() {
     assert.success().stdout("foobar\n");
 }
 
+#[test]
+fn it_cut_using_multibyte_delimiters() {
+    let mut cmd = Command::cargo_bin(env!("CARGO_PKG_NAME")).unwrap();
+
+    let assert = cmd
+        .args(["-d", "...", "-f", "2"])
+        .write_stdin("foo...bar")
+        .assert();
+
+    assert.success().stdout("bar\n");
+}
+
 #[test]
 fn it_works_on_multiple_lines() {
     let mut cmd = Command::cargo_bin(env!("CARGO_PKG_NAME")).unwrap();
 
     let assert = cmd
         .args(["-f", "2"])
-        .write_stdin("hello\nfoobar\tbaz")
+        .write_stdin("hello\tworld\nfoobar\tbaz")
         .assert();
 
-    assert.success().stdout("hello\nbaz\n");
+    assert.success().stdout("world\nbaz\n");
 }
 
 #[test]
@@ -87,6 +180,7 @@ fn it_compresses_delimiters_when_requested_and_handles_boundaries() {
     assert.success().stdout("-foo-bar-\n");
 }
 
+#[cfg(feature = "regex")]
 #[test]
 fn it_cuts_on_characters() {
     let mut cmd = Command::cargo_bin(env!("CARGO_PKG_NAME")).unwrap();
@@ -99,6 +193,19 @@ fn it_cuts_on_characters() {
     assert.success().stdout("🤩😝\n");
 }
 
+#[cfg(not(feature = "regex"))]
+#[test]
+fn it_fails_to_cut_characters_when_no_regex_support() {
+    let mut cmd = Command::cargo_bin(env!("CARGO_PKG_NAME")).unwrap();
+
+    let assert = cmd
+        .args(["--characters", "2,-2"])
+        .write_stdin("😁🤩😝😎")
+        .assert();
+
+    assert.failure().stderr("tuc: runtime error. The use of --characters requires `tuc` to be compiled with `regex` support\n");
+}
+
 #[test]
 fn it_cuts_on_bytes() {
     let mut cmd = Command::cargo_bin(env!("CARGO_PKG_NAME")).unwrap();
@@ -137,7 +244,7 @@ fn it_can_complement_the_fields() {
         .write_stdin("a b c")
         .assert();
 
-    assert.success().stdout("\n");
+    assert.failure().stderr("Error: the complement is empty\n");
 }
 
 #[test]
@@ -238,13 +345,27 @@ fn it_format_fields() {
     let mut cmd = Command::cargo_bin(env!("CARGO_PKG_NAME")).unwrap();
 
     let assert = cmd
-        .args(["-f", "Say {1} to our {2}.\nJust {{saying}}"])
+        .args(["-d", ",", "-f", "f1: {1}, f2: {2}"])
+        .write_stdin("hello,world")
+        .assert();
+
+    assert.success().stdout("f1: hello, f2: world\n");
+
+    let mut cmd = Command::cargo_bin(env!("CARGO_PKG_NAME")).unwrap();
+
+    let assert = cmd
+        .args([
+            "--fallback-oob",
+            "generic fallback",
+            "-f",
+            "Say {1} to our {2}.\nJust {{saying}} {3=dedicated fallback} {4}",
+        ])
         .write_stdin("hello\tworld")
         .assert();
 
     assert
         .success()
-        .stdout("Say hello to our world.\nJust {saying}\n");
+        .stdout("Say hello to our world.\nJust {saying} dedicated fallback generic fallback\n");
 }
 
 #[test]
@@ -318,7 +439,7 @@ fn it_fails_if_there_are_unknown_arguments() {
     let assert = cmd.args(["--whatever"]).write_stdin("foobar").assert();
 
     assert.failure().stderr(
-        "tuc: unexpected arguments [\"--whatever\"]\nTry 'tuc --help' for more information.\n",
+        "tuc: unexpected arguments: [\"--whatever\"]\nTry 'tuc --help' for more information.\n",
     );
 }
 
@@ -358,6 +479,7 @@ fn it_fails_when_the_regex_is_malformed() {
         "tuc: runtime error. The regular expression is malformed.",
     ));
 
+    let mut cmd = Command::cargo_bin(env!("CARGO_PKG_NAME")).unwrap();
     let assert = cmd.args(["-e", "[", "-g"]).write_stdin("foobar").assert();
 
     assert.failure().stderr(predicates::str::starts_with(
@@ -373,7 +495,7 @@ fn does_not_panic_if_attemtping_to_use_regex_arg_with_noregex_build() {
     let assert = cmd.args(["-e", "."]).write_stdin("foobar").assert();
 
     assert.failure().stderr(
-        "tuc: unexpected arguments [\"-e\", \".\"]\nTry 'tuc --help' for more information.\n",
+        "tuc: unexpected arguments: [\"-e\", \".\"]\nTry 'tuc --help' for more information.\n",
     );
 }
 
@@ -392,6 +514,7 @@ fn it_emit_output_as_json() {
     );
 }
 
+#[cfg(feature = "regex")]
 #[test]
 fn it_emit_output_as_json_even_when_cutting_on_chars() {
     let mut cmd = Command::cargo_bin(env!("CARGO_PKG_NAME")).unwrap();
@@ -418,6 +541,17 @@ fn it_does_not_allow_to_replace_delimiter_with_json() {
         .stderr("tuc: runtime error. The use of --replace with --json is not supported\n");
 }
 
+#[test]
+fn it_fails_if_you_use_json_and_nojoin_at_once() {
+    let mut cmd = Command::cargo_bin(env!("CARGO_PKG_NAME")).unwrap();
+
+    let assert = cmd.args(["--json", "--no-join"]).assert();
+
+    assert
+        .failure()
+        .stderr("tuc: runtime error. Using both --json and --no-join is not permitted\n");
+}
+
 #[test]
 fn it_is_not_allowed_to_use_character_with_nojoin() {
     let mut cmd = Command::cargo_bin(env!("CARGO_PKG_NAME")).unwrap();
@@ -429,6 +563,18 @@ fn it_is_not_allowed_to_use_character_with_nojoin() {
     );
 }
 
+#[cfg(not(feature = "regex"))]
+#[test]
+fn it_cannot_use_characters_without_regex() {
+    let mut cmd = Command::cargo_bin(env!("CARGO_PKG_NAME")).unwrap();
+
+    let assert = cmd.args(["-c", "1"]).assert();
+
+    assert.failure().stderr(
+        "tuc: runtime error. The use of --characters requires `tuc` to be compiled with `regex` support\n",
+    );
+}
+
 #[test]
 fn it_does_not_support_json_on_lines() {
     let mut cmd = Command::cargo_bin(env!("CARGO_PKG_NAME")).unwrap();
@@ -461,3 +607,98 @@ fn it_cannot_format_fields_alongside_json() {
         .failure()
         .stderr("tuc: runtime error. Cannot format fields when using --json\n");
 }
+
+#[test]
+fn it_accept_fixed_memory_option() {
+    let mut cmd = Command::cargo_bin(env!("CARGO_PKG_NAME")).unwrap();
+
+    let assert = cmd
+        .args(["-d", "-", "-f", "2", "-M", "1"])
+        .write_stdin("ab-cd")
+        .assert();
+
+    assert.success().stdout("cd\n");
+}
+
+#[test]
+fn it_fails_if_fixed_memory_option_is_used_with_unsupported_fields() {
+    let mut cmd = Command::cargo_bin(env!("CARGO_PKG_NAME")).unwrap();
+
+    let assert = cmd
+        .args(["-d", "-", "-f", "2", "-M", "1", "--json"])
+        .write_stdin("ab-cd")
+        .assert();
+
+    assert
+        .failure()
+        .stderr("tuc: runtime error. StreamOpt supports solely forward fields, join and single-character delimiters\n");
+}
+
+#[test]
+fn it_fails_if_fixed_memory_value_is_zero() {
+    let mut cmd = Command::cargo_bin(env!("CARGO_PKG_NAME")).unwrap();
+
+    let assert = cmd
+        .args(["-d", "-", "-f", "2", "-M", "0"])
+        .write_stdin("ab-cd")
+        .assert();
+
+    assert
+        .failure()
+        .stderr("tuc: runtime error. --fixed-memory cannot be 0\n");
+}
+
+#[test]
+fn it_checks_possible_regression_on_negative_indices_and_fallbacks() {
+    let mut cmd = Command::cargo_bin(env!("CARGO_PKG_NAME")).unwrap();
+
+    let assert = cmd
+        .args(["-d", "-", "-f", "{-3=hello}"])
+        .write_stdin("a-b-c\nd-e")
+        .assert();
+
+    assert.success().stdout("a\nhello\n");
+}
+
+#[cfg(feature = "regex")]
+#[test]
+fn it_checks_possible_regression_on_negative_indices_and_fallbacks_with_regex() {
+    let mut cmd = Command::cargo_bin(env!("CARGO_PKG_NAME")).unwrap();
+
+    let assert = cmd
+        .args(["-e", "-", "-f", "{-3=hello}"])
+        .write_stdin("a-b-c\nd-e")
+        .assert();
+
+    assert.success().stdout("a\nhello\n");
+}
+
+#[test]
+fn it_checks_possible_regression_fallbacks_single_delimiter() {
+    let mut cmd = Command::cargo_bin(env!("CARGO_PKG_NAME")).unwrap();
+
+    let assert = cmd
+        .args(["-d", "-", "-f", "{2=hello},{3}"])
+        .write_stdin("a-b-c\nd")
+        .assert();
+
+    assert
+        .failure()
+        .stdout("b,c\nhello,")
+        .stderr("Error: Out of bounds: 3\n");
+}
+
+#[test]
+fn it_checks_possible_regression_fallbacks_multi_delimiter() {
+    let mut cmd = Command::cargo_bin(env!("CARGO_PKG_NAME")).unwrap();
+
+    let assert = cmd
+        .args(["-d", "--", "-f", "{2=hello},{3}"])
+        .write_stdin("a--b--c\nd")
+        .assert();
+
+    assert
+        .failure()
+        .stdout("b,c\nhello,")
+        .stderr("Error: Out of bounds: 3\n");
+}
